// ======================================================================= //
// Class Commentator
// file : commentator.inl
// (c) copyright Dumas & Saunders 1999
// ======================================================================= //

#include <math.h>
#include <iostream.h>
#include "commentator.h"

// -----------------------------------------------------
// Mathematical routines
// -----------------------------------------------------
#ifndef GIVMAX
#define GIVMAX(a,b) ((a)<(b) ? b:a)
#endif

#ifndef __GIVMATHS__
#define __GIVMATHS__


#define ispolynomial(c,a,k1,t1,k2,t2) { \
    (a) = log((double)(t2)/(double)(t1)) / log((double)(k2)/(double)(k1)); \
    (c) = ( (t2)/exp((a)*log((double)(k2))) + (t1)/exp((a)*log((double)(k1))) )/2.0; \
}

#define polynomialize(k3,t3,c,a) { (t3) = (c)*exp((a)*log(k3)); } 


#endif
// -------------------------------
// Inline members
// -------------------------------
inline Commentator::Commentator(long gt, long gh, long tab = 3, long depth = 5)
{
    gapTimingLevel = gt;
    gapHomologyLevel = gh;
    indentationStep = tab;
    estimatesDepth = GIVMAX(depth,2);
}

inline void Commentator::start(const char* id, const char* msg, long msglevel, long msgclass)
{  
    IDs.push(id);
    report(msg, msglevel, msgclass);
    Timer activ; activ.clear(); activ.start();
    Timers.push( activ );
    Estimator est; 
    Estimates.push( est );
}

inline void Commentator::stop(const char* msg, long msglevel, long msgclass) 
{
    (Timers.top()).stop();
    if (msg != "")
        report(msg, msglevel, msgclass);
    if (header(msglevel, TIMING_MEASURE)) {
        cerr << "cpu  time : " << (Timers.top()).usertime() << " seconds" << endl;
        if (IDs.size() == 1) {
            header(msglevel, TIMING_MEASURE);
            cerr << "sys  time : " << (Timers.top()).systime() << " seconds" << endl;
            header(msglevel, TIMING_MEASURE);
            cerr << "real time : " << (Timers.top()).realtime() << " seconds" << endl;
        }
    }
    IDs.pop();
    Timers.pop();
    Estimates.pop();
}

inline void Commentator::printheader(char information) const 
{
        long indent = indentationStep*IDs.size();
        cerr << "#I:" << information;
        for (long i = 0; i < indent; i++) cerr << " ";
        if (IDs.size()) cerr << IDs.top() << ": ";
        else cerr << " ";
}

inline long Commentator::header(long msglevel, long msgclass) const 
{
    long isp; char information;
    if (isp = intern_printed(information, msglevel, msgclass))
        printheader(information);
    return isp;
}




inline void Commentator::report(const char* msg, long msglevel, long msgclass) const
{
    if (header(msglevel, msgclass))
        cerr << msg << endl;
}

inline void Commentator::progress(const char* msg, long msglevel, long k, long n)
{

    if ((k) && ( header( msglevel, TIMING_ESTIMATE) ) ) {
        Timer star = Timers.top();
        (Timers.top()).stop();
        double t3 = (Timers.top()).usertime();
        Estimates.top().push_back( StepsAndTime( k, t3) );
        Estimator::iterator Etop = Estimates.top().begin();

        double t1 = (*Etop).gettime();
        long k1 = (*Etop).getsteps();

        if (Estimates.top().size() >= 3) {

            double t2 = (*(++Etop)).gettime();
            long k2 = (*Etop).getsteps();

            double ce, ae;
            ispolynomial(ce, ae, k1, t1, k2, t2);
            double guess;
            if (Estimates.top().size() >= 4) {
                double t2p = (*(++Etop)).gettime();
                long k2p = (*Etop).getsteps();
                double cep, aep;
                ispolynomial(cep, aep, k1, t1, k2p, t2p);
                double guess2, guess2p;
                polynomialize(k, guess2, ce, ae);
                polynomialize(k, guess2p, cep, aep);
                guess2 = (t3 - guess2); if (guess2 < 0) guess2 = - guess2;
                guess2p = (t3 - guess2p); if (guess2p < 0) guess2p = - guess2p;
                if (guess2 > guess2p) {
                    ce = cep;
                    ae = aep;
                }
            }
            
            polynomialize(n,guess,ce,ae); 
            guess = (t3 - guess); if (guess < 0) guess = - guess;

            cerr << k << " out of " << n << " " << msg << " done; Elapsed:" << t3 << "; " << guess << " seconds to expected completion." << endl;
        } else {
            cerr << k << " out of " << n << " " << msg << " done; " << endl;
        }
        

        if (Estimates.top().size() > estimatesDepth)
            Estimates.top().pop_front();
        Timers.top() = star;
    }
}



inline long Commentator::intern_printed(char& information, long msglevel, long msgclass) const 
{
    if (msglevel == -1) {
        information = 'U';
        return 1;
    }

//     if (msgclass == INTERNAL_ERROR) {
//         information = 'E';
//         return 1;
//     }
//     if (msgclass == INTERNAL_WARNING) {
//         information = 'W';
//         return gapHomologyLevel > msglevel;
//     }
//     if (msgclass == PARTIAL_RESULT) {
//         information = 'R';
//         return gapHomologyLevel > msglevel + IDs.size() - 1;
//     }
//     if (msgclass == INTERNAL_DESCRIPTION) {
//         information = 'D';
//         return gapHomologyLevel > msglevel + IDs.size();
//     }
//     if (msgclass == TIMING_MEASURE) {
//         information = 'T';
//         return gapTimingLevel > msglevel + IDs.size() - 1;
//     }
//     if (msgclass == TIMING_ESTIMATE) {
//         information = 'S';
//         return gapTimingLevel > msglevel + IDs.size();
//     }

    switch(msgclass) {
        case INTERNAL_ERROR:
            information = 'E';
            return 1;
        case INTERNAL_WARNING:
            information = 'W';
            return gapHomologyLevel > msglevel;
        case PARTIAL_RESULT:
            information = 'R';
            return gapHomologyLevel > msglevel + IDs.size() - 1;
        case INTERNAL_DESCRIPTION:
            information = 'D';
            return gapHomologyLevel > msglevel + IDs.size();
        case TIMING_MEASURE:
            information = 'T';
            return gapTimingLevel > msglevel + IDs.size() - 1;
        case TIMING_ESTIMATE:
            information = 'S';
            return gapTimingLevel > msglevel + IDs.size();
        default:
            information = 'U';
            return 0;
    }
    
}

inline long Commentator::printed(long msglevel, long msgclass) const 
{
    char c; return intern_printed(c, msglevel, msgclass);
}
